#!/usr/bin/python
# -*- coding: ascii -*-
"""
CRCx implementation.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

from typing import List
from embutils.utils.bytes import as_hex, bitmask, reverse_bits


class CRC:
    """Table driven CRC calculation.
    """
    def __init__(self,
                 name: str = "CRC16_CCITT_FALSE", size: int = 16,
                 poly: int = 0x1021, crc_init: int = 0xFFFF, xor_out: int = 0x0000,
                 rev_in: bool = False, rev_out: bool = False
                 ):
        """Class initialization.

        Notes:
            - For the moment this implementation don't support CRCs with size < 8bits
            - The CRC configuration values cant be changed after initialization.
            - The CRC configuration values will be adjusted to the selected width: CRC16 poly 0x101021 -> 0x1021.

        Args:
            name    (str) : Identification name.
            size    (int) : Size in bits.
            poly    (int) : Generator polynomial value.
            crc_init(int) : Value used to initialize the CRC value.
            xor_out (int) : Value to be xored to the CRC sum before being returned.
            rev_in  (bool): If true every input byte will be reversed before being feed to the calculation.
            rev_out (bool): If true the final CRC value will be reversed before being returned.
        """
        # Store configurations...
        self._name = name
        self._size = size
        self._rev_in = rev_in
        self._rev_out = rev_out

        # Adjust the values using a bitmask
        self._mask = bitmask(bit=(size - 1), fill=True)
        self._poly = self._mask & poly
        self._xor_out = self._mask & xor_out
        self._crc_init = self._mask & crc_init

        # Generate the lookup table
        self._table = self._gen_table()

    def __repr__(self) -> str:
        """Get the class representation string.

        Return:
            str: Class representation string.
        """
        msg = f'{self.__class__.__name__}(name={self.name}, size={self.size}, ' +\
              f'poly={self.poly_hex}, crc_in={self.crc_init_hex}, xor_out={self.xor_out_hex}, ' +\
              f'ref_in={self.reverse_in}, ref_out={self.reverse_out})'
        return msg

    # Properties ================================
    @property
    def name(self) -> str:
        """CRC name.

        Returns:
            str: CRC algorithm name.
        """
        return self._name

    @property
    def size(self) -> int:
        """CRC bit size.

        Returns:
            int: CRC size (in bits).
        """
        return self._size

    @property
    def poly(self) -> int:
        """CRC polynomial.

        Returns:
            int: CRC polynomial value.
        """
        return self._poly

    @property
    def poly_hex(self) -> str:
        """CRC polynomial value (as hex string).

        Returns:
            str: CRC polynomial value in hex.
        """
        return as_hex(data=self._poly, size=self._size)

    @property
    def xor_out(self) -> int:
        """CRC out XOR value.

        Returns:
            int: CRC output XOR value.
        """
        return self._xor_out

    @property
    def xor_out_hex(self) -> str:
        """CRC out XOR value (as hex string).

        Returns:
            str: CRC output XOR value in hex.
        """
        return as_hex(data=self._xor_out, size=self._size)

    @property
    def crc_init(self) -> int:
        """CRC initial sum value.

        Returns:
            int: CRC input value.
        """
        return self._crc_init

    @property
    def crc_init_hex(self) -> str:
        """CRC initial sum value (as hex string).

        Returns:
            str: CRC input value in hex.
        """
        return as_hex(data=self._crc_init, size=self._size)

    @property
    def reverse_in(self) -> bool:
        """Reverse input data bits.

        Returns:
            bool: True if the bits are being reversed.
        """
        return self._rev_in

    @property
    def reverse_out(self) -> bool:
        """Reverse CRC output bits.

        Returns:
            bool: True if the bits are being reversed.
        """
        return self._rev_out

    @property
    def lookup_table(self) -> List[int]:
        """Return the CRC lookup table.

        Returns:
            list: CRC lookup table.
        """
        return self._table

    # API =======================================
    def compute(self, data: bytearray, crc_init: int = None) -> int:
        """Compute the CRC of the given data.

        Args:
            data (bytearray): Bytes to compute the CRC over.
            crc_init (int): Alternative initial value.

        Returns:
            int: CRC value.
        """
        # Select the initial value
        crc_init = self._crc_init if crc_init is None else crc_init

        # Process depending on size
        if self._size >= 8:
            # Define required constants
            shift = self._size - 8

            # Set the initial value and iterate through bytes
            crc = crc_init
            for byte in data:
                # Get the byte (reverse if necessary)
                byte = reverse_bits(data=byte, size=8) if self._rev_in else byte
                # Get index and update CRC
                pos = 0xFF & ((crc >> shift) ^ byte)
                crc = self._mask & ((crc << 8) ^ self._table[pos])

        else:
            # Define required constants
            shift = 8 - self._size

            if self._rev_in:
                # Set the initial value and iterate through bytes
                crc = reverse_bits(data=crc_init, size=self._size)
                for byte in data:
                    # Get index and update CRC
                    pos = 0xFF & (crc ^ byte)
                    crc = self._mask & ((crc >> 8) ^ self._table[pos])
                # Adjust output value
                crc = reverse_bits(data=crc, size=self._size)

            else:
                # Set the initial value and iterate through bytes
                crc = crc_init << shift
                for byte in data:
                    # Get index and update CRC
                    pos = 0xFF & (crc ^ byte)
                    crc = (self._mask << shift) & ((crc << self._size) ^ (self._table[pos] << shift))
                # Adjust output value
                crc >>= shift

        # Prepare the final CRC value and apply XOR
        crc = reverse_bits(data=crc, size=self._size) if self._rev_out else crc
        return crc ^ self._xor_out

    # API Private ===============================
    def _gen_table(self) -> List[int]:
        """Generate the lookup table for fast CRC computation.

        Returns:
            List[int]: Table with 256 pre-computed CRC values.
        """
        # Generate the lookup table
        out = []
        if self._size >= 8:
            # Define required constants
            shift = self._size - 8
            check = bitmask(bit=(self._size - 1))

            # Iterate through all the table values
            for idx in range(256):
                # The current byte is the table index
                byte = idx << shift

                # Compute the CRC value
                for bit in range(8):
                    byte = ((byte << 1) ^ self._poly) if ((byte & check) != 0) else (byte << 1)

                # Store the value on the table
                out.append(self._mask & byte)

        else:
            # Define required constants
            shift = (8 - self._size)
            check = bitmask(bit=7)

            # Iterate through all the table values
            for idx in range(256):
                # The current byte is the table index
                byte = reverse_bits(data=idx, size=8) if self._rev_in else idx

                # Compute the CRC value
                for bit in range(8):
                    byte = ((byte << 1) ^ (self._poly << shift)) if ((byte & check) != 0) else (byte << 1)

                # Store the value on the table
                byte = (reverse_bits(data=(byte >> shift), size=self._size) << shift) if self._rev_in else byte
                out.append(self._mask & (byte >> shift))

        # Return the lookup table
        return out



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Enumeration utilities.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

from enum import IntEnum


class IntEnumMod(IntEnum):
    """Modified IntEnum class.
    """
    def __repr__(self) -> str:
        """Get the class representation string.

        Return:
            str: Class representation string.
        """
        return f'{self.__class__.__name__}({self.__str__()})'

    def __str__(self) -> str:
        """Get the value as string.

        Return:
            str: Enum value as string.
        """
        return f'{self.name}(0x{self.value:X})'

    @classmethod
    def from_int(cls, value: int) -> 'IntEnumMod':
        """If available, convert the input to a enum object.

        Args:
            value (int): Item to be converted.

        Returns:
            IntEnumMod: Enumeration object.
        """
        if cls.has_value(value=value):
            return cls(value)
        else:
            raise ValueError("Value 0x{:02X} is not defined.".format(value))

    @classmethod
    def has_value(cls, value: int) -> bool:
        """Get if the given value exists on the class.

        Args:
            value (int): Item to check.

        Returns:
            bool: True if exists, false otherwise.
        """
        return value in cls._value2member_map_



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serial interface implementation.
This class represents the point closer to the user. Here the DEV has to
define/use payload/commands/frame layouts to communicate with the device.

NOTE: All the functions that interacts with the device (ex: read_serial_number)
should be defined here.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

import time
from embutils.serial.data import Frame, FrameHandler
from embutils.serial.core import FrameStream, SerialDevice
from embutils.utils import time_elapsed, EventHook, LOG_SDK, UsbID
from threading import Lock
from typing import Union


logger_sdk = LOG_SDK.logger


class SerialInterface:
    """Serial command interface implementation.
    This class implement all serial command wrappers and handlers.

    The available events are:
        1. on_port_reconnect: This event is emitted when the port is able to reconnect.
            Subscribe using callback with syntax:
                def <callback>()

        2. on_port_disconnect: This event is emitted when the port get disconnected.
            Subscribe using callback with syntax:
                def <callback>()

        3. on_frame_received: This event is emitted when a frame is received.
            Subscribe using callback with syntax:
                def <callback>(frame: Frame)
        """
    # Command wait response timeout
    RESPONSE_TIMEOUT_S = 0.5

    # Serial device settings
    SERIAL_SETTINGS = {
        'baudrate': 230400,
        'bytesize': 8,
        'stopbits': 1,
        'parity':   'N',
        'timeout':  0.1
        }

    def __init__(self, frame_handler: FrameHandler, port: str = None, looped: bool = False) -> None:
        """Class initialization.

        Args:
            frame_handler (FrameHandler): Handler used to process incoming frames.
            port (str): Serial device port.
            looped (bool): Enable test mode (looped serial).
        """
        # Add lock for multithreading usage
        self._lock = Lock()

        # Response timeout configuration
        self._timeout = self.RESPONSE_TIMEOUT_S

        # Public events
        self.on_port_reconnect = EventHook()
        self.on_port_disconnect = EventHook()
        self.on_frame_received = EventHook()

        # Frame Stream: Communications thread
        # Start serial device
        if looped:
            serial = SerialDevice(usb_id=UsbID(vid=0xDEAD, pid=0xBEEF), settings=self.SERIAL_SETTINGS, looped=True)
        else:
            serial = SerialDevice(port=port, settings=self.SERIAL_SETTINGS)

        # Initialize frame stream and attach callbacks
        self._frame_stream = FrameStream(serial_device=serial, frame_handler=frame_handler)
        self._frame_stream.on_port_reconnect  += self.on_port_reconnect.emit
        self._frame_stream.on_port_disconnect += self.on_port_disconnect.emit
        self._frame_stream.on_frame_received  += self._process_frame

        logger_sdk.info('Interface initialized on: {}'.format(self._frame_stream.serial_device))

    @property
    def lock(self) -> Lock:
        """Return the Command Interface lock for multithreading
        support.

        Return:
            Lock: Mutex.
        """
        return self._lock

    @property
    def frame_stream(self) -> FrameStream:
        """Return the current frame stream handler.

        Return:
            FrameStream: Handler used to send/receive frames over serial.
        """
        return self._frame_stream

    @property
    def timeout(self) -> float:
        """Returns the current response timeout.

        Return:
            float: Time in [s].
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value: float) -> None:
        """Set the current response timeout.

        Args:
            value (float): Time in [s].
        """
        if value <= 0.0:
            raise ValueError('The response timeout needs to be greater than zero.')
        self._timeout = value

    def _process_frame(self, frame: Frame) -> None:
        """Method used to handle what to do when a frame is received.

        Args:
            frame (Frame): Frame to be processed.
        """
        # Do special actions:

        # Pass frame to user:
        self.on_frame_received.emit(frame=frame)

    def transmit(self, send: Frame, resp_logic: callable = None, resp_timeout: float = None) -> Union[None, Frame]:
        """Send a frame and wait for the response based on custom logic.
        The response detection callback syntax is:
            def <callback>(sent: Frame, recv: Frame)

        Args:
            send (Frame): Frame to be sent to the device.
            resp_logic (callable, optional): Response detection logic. The response detection is disabled when None.
            resp_timeout (float, optional): Response timeout. Use default interface timeout when None.

        Returns:
            Union[None, Frame]: None if timeout of no response detection is requested. If response detected Frame.
        """
        # SEND ------------------------
        # Prepare and send frame
        logger_sdk.debug('Sending frame...')
        self._frame_stream.send_frame(frame=send)

        # RECEIVE ---------------------
        # Return nothing if we don't need response
        if resp_logic is None:
            logger_sdk.debug('Response is not needed.')
            return None

        # Process wait response
        logger_sdk.debug('Waiting for frame response...')
        resp   = None
        ready  = False
        resp_timeout = self._timeout if (resp_timeout is None) else resp_timeout

        # Receive logic callback
        def on_frame_received(frame: Frame):
            nonlocal send, resp, resp_logic, ready
            ready = resp_logic(sent=send, recv=frame)
            if ready:
                resp = frame

        # Perform wait response logic
        self.on_frame_received += on_frame_received
        tm_start = time.time()
        while not ready and (time_elapsed(tm_start) < resp_timeout):
            time.sleep(0.01)
        self.on_frame_received -= on_frame_received

        # Check data
        logger_sdk.debug('Frame response: {state} after {elapsed:.03f}[s]'.format(
            state='Received' if ready else 'Timeout',
            elapsed=time_elapsed(start=tm_start)
            ))
        return resp

    def stop(self) -> None:
        """Stops the frame stream process.
        """
        self._frame_stream.stop()

    def join(self) -> None:
        """Maintains the process alive while the interface is working.
        """
        while self._frame_stream.is_alive:
            time.sleep(0.5)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serialized object definition.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

from abc import abstractmethod


class Serialized:
    """Serial item abstract definition.
    """
    LENGTH = 0

    @property
    def length(self) -> int:
        """Length of the payload contents.

        Returns:
            int: Payload length
        """
        return self.LENGTH

    @abstractmethod
    def serialize(self) -> bytearray:
        """Serialize the item to a bytearray.

        Returns:
            bytearray: Serialized object.
        """
        pass

    @staticmethod
    @abstractmethod
    def deserialize(data: bytearray) -> 'Serialized':
        """Deserializes an item form the input bytes.

        Args:
            data (bytearray): Input byte stream.

        Returns:
            SerialItem: Deserialized item.
        """
        pass



#!/usr/bin/python
# -*- coding: ascii -*-
"""
USB device ID.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

import re
from typing import Tuple
from embutils.utils.bytes import bitmask


class UsbID(int):
    """USB ID implementation.
    This class represent the USB VID/PID values.
    """
    ID_MASK     = bitmask(bit=(16 - 1), fill=True)
    ID_PATTERN  = '(?:[0-9a-f]{4}):(?:[0-9a-f]{4})'

    def __new__(cls, vid: int = 0x0000, pid: int = 0x0000) -> 'UsbID':
        """Generate a new integer item with the USB ID values.

        Args:
            vid (int): Vendor ID.
            pid (int): Product ID.

        Returns:
            UsbID: USB ID instance.
        """
        # Define values
        uid = UsbID._pack(vid=vid, pid=pid)
        _vid, _pid = UsbID._unpack(value=uid)

        # Generate item
        obj = int.__new__(cls, uid)
        obj._vid = _vid
        obj._pid = _pid
        return obj

    def __repr__(self) -> str:
        """Get the class representation string.

        Return:
            str: Class representation string.
        """
        return f'{self.__class__.__name__}(vid=0x{self.vid:04X}, pid=0x{self.pid:04X})'

    def __str__(self) -> str:
        """Get the class as a string.

        Return:
            str: Class string.
        """
        return 'VID:PID={:04X}:{:04X}'.format(self._vid, self._pid)

    @property
    def vid(self) -> int:
        """Get the VID value.

        Return:
            int: VID value.
        """
        return self._vid

    @property
    def pid(self) -> int:
        """Get the PID value.

        Return:
            int: PID value.
        """
        return self._pid

    @property
    def vid_as_hex(self) -> str:
        """Get the VID value as hex.

        Returns:
            str: String with the hex value.
        """
        return '0x{:02X}'.format(self._vid)

    @property
    def pid_as_hex(self) -> str:
        """Get the PID value as hex.

        Returns:
            str: String with the hex value.
        """
        return '0x{:02X}'.format(self._pid)

    @staticmethod
    def from_int(uid: int) -> 'UsbID':
        """Unpacks the USB ID from a integer.

        Args:
            uid (int): USB ID (based on VID/PID).

        Returns:
            UsbID: Object with VID/PID based on input.
        """
        vid, pid = UsbID._unpack(value=uid)
        return UsbID(vid=vid, pid=pid)

    @staticmethod
    def from_str(uid: str) -> 'UsbID':
        """Unpacks the USB ID from a string.

        Args:
            uid (str): USB ID (based on VID/PID) with format VID:PID.

        Returns:
            UsbID: Object with VID/PID based on input.
        """
        match = re.findall(pattern=UsbID.ID_PATTERN, string=uid.lower())
        if match:
            return UsbID.from_int(uid=int(match[0].replace(':', ''), 16))
        raise ValueError('Invalid USB ID: {}'.format(uid))

    @staticmethod
    def _pack(vid: int, pid: int) -> int:
        """Pack the VID and UID values into an int.

        Args:
            vid (int): Vendor ID.
            pid (int): Product ID.

        Returns:
            int: Combined VID/PID value.
        """
        pid = UsbID.ID_MASK & pid
        vid = UsbID.ID_MASK & vid
        return (vid << 16) | pid

    @staticmethod
    def _unpack(value: int) -> Tuple[int, int]:
        """Unpack the VID and PID values from an int.

        Args:
            value (int): Combined VID/PID value.

        Returns:
            int, int: VID, PID values.
        """
        pid = UsbID.ID_MASK & value
        vid = UsbID.ID_MASK & (value >> 16)
        return vid, pid



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Frame stream implementation.
The frame stream is the class used to read/write frames from/to the serial
device.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

import time
from embutils.serial.data import Frame, FrameHandler
from embutils.serial.core.serial_device import SerialDevice
from embutils.utils import EventHook, LOG_SDK, ThreadItem


logger_sdk = LOG_SDK.logger


class FrameStream:
    """Frame stream implementation.
    This class is used to send and receive frames through the serial device
    in an asynchronous way. When a new frame is received this class will notify
    using events.

    The available events are:
        1. on_frame_received: This event is emitted when a frame is received.
            Subscribe using callback with syntax:
                def <callback>(frame: Frame)

        2. on_port_reconnect: This event is emitted when the port is able to reconnect.
            Subscribe using callback with syntax:
                def <callback>()

        3. on_port_disconnect: This event is emitted when the port get disconnected.
            Subscribe using callback with syntax:
                def <callback>()
    """
    def __init__(self, serial_device: SerialDevice, frame_handler: FrameHandler) -> None:
        """Class initialization.

        Args:
            serial_device (SerialDevice): Used to read/write from serial.
            frame_handler (FrameHandler): Used to interpret the incoming bytes as a frame.
        """
        # Initialize
        self._serial_device = serial_device
        self._frame_handler = frame_handler

        # Events
        self.on_frame_received = EventHook()
        self.on_port_reconnect = EventHook()
        self.on_port_disconnect = EventHook()

        # Debug prints for received
        self.on_frame_received += lambda frame: self._print_debug(frame=frame, received=True)

        # Thread related
        self._is_active = True
        self._is_paused = False
        self._pause_active = False
        self._thread = ThreadItem(name=self.__class__.__name__, target=self._process)

    def __del__(self) -> None:
        """Class destructor. Stop the thread and remove the associated serial device.
        """
        if self._is_active:
            self.stop()
        del self._serial_device

    @property
    def serial_device(self) -> SerialDevice:
        """Get the current serial device handler.

        Returns:
            SerialDevice: Handler used to interact with the serial port.
        """
        return self._serial_device

    @serial_device.setter
    def serial_device(self, handler: SerialDevice) -> None:
        """Configure a new handler for the serial device.

        NOTE: If the new handler has no port defined the current device
        will be preserved.

        Args:
            handler (SerialDevice): Serial device handler.
        """
        self.pause()
        if handler.port is not None:
            self._serial_device = handler
        self.resume()

    @property
    def frame_handler(self) -> FrameHandler:
        """Get the current frame handler.

        Returns:
            FrameHandler: Handler used to process the received bytes into a frame.
        """
        return self._frame_handler

    @frame_handler.setter
    def frame_handler(self, handler: FrameHandler) -> None:
        """Configure a new frame handler.

        Args:
            handler (FrameHandler): Frame handler.
        """
        self.pause()
        if handler:
            self._frame_handler = handler
        self.resume()

    @property
    def is_alive(self) -> bool:
        """Return if the thread is alive.

        Returns:
            bool: True if alive, false otherwise.
        """
        return self._thread.is_alive()

    @property
    def is_working(self) -> bool:
        """Return if the process is being executed.

        Returns;
            bool: True if working, false if stopped or paused.
        """
        return self.is_alive and self._serial_device.is_open and not self._is_paused

    def send_frame(self, frame: Frame) -> None:
        """Send the specified frame using the serial device.

        Args:
            frame (Frame): Frame to be sent.
        """
        if self.is_working:
            self._print_debug(frame=frame, received=False)
            self._serial_device.write(data=frame.serialize())

    def resume(self) -> None:
        """Resume the data transmission.
        """
        if self.is_alive and self._serial_device.open():
            self._serial_device.flush()
            self._is_paused = False
            self._pause_active = False
            logger_sdk.info("Stream resumed.")

    def pause(self) -> None:
        """Pause the data transmission.
        """
        if self.is_working:
            self._is_paused = True
            while not self._pause_active:
                time.sleep(0.01)
            self._serial_device.close()
            logger_sdk.info("Stream paused.")

    def stop(self) -> None:
        """Stop the transmission and kill the thread.
        """
        self._is_active = False
        while self._thread.is_alive():
            time.sleep(0.01)
        self._serial_device.close()
        logger_sdk.info("Stream stopped.")

    def _process(self) -> None:
        """Frame transmission main loop.
        This loop handle the frame reception:
            1. Read the FrameHandler required bytes.
            2. Send the bytes to be processed by the FrameHandler.
            3. Handle disconnection status.
        """
        logger_sdk.info("Frame stream started.")

        # Do this periodically
        while self._is_active:
            # Pause state
            if self._is_paused:
                if not self._pause_active:
                    self._pause_active = True
                time.sleep(0.01)
                continue

            # Receive and process data
            if not self._frame_handler.read_process(serial_device=self._serial_device, emitter=self.on_frame_received):
                # Device disconnected...
                logger_sdk.info("Device disconnected: {}".format(self._serial_device))
                ThreadItem(
                    name='{}.{}'.format(self.__class__.__name__, 'on_port_disconnect'),
                    target=self.on_port_disconnect.emit
                    )
                if self._reconnect():
                    ThreadItem(
                        name='{}.{}'.format(self.__class__.__name__, 'on_port_reconnect'),
                        target=self.on_port_reconnect.emit
                        )

            # Give some time
            time.sleep(0.01)

    def _reconnect(self) -> bool:
        """Start a reconnection attempt to the serial device.

        Returns:
            bool: True if reconnection succeeded, false otherwise.
        """
        status = False
        self._serial_device.close()
        logger_sdk.info("Starting reconnection attempt on {}".format(self._serial_device))
        while self._is_active:
            if self._serial_device.open():
                logger_sdk.info("Device {} reconnected.".format(self._serial_device))
                status = True
                break
            else:
                logger_sdk.info("Reconnection attempt on {} failed.".format(self._serial_device))
                time.sleep(0.5)
        return status

    @staticmethod
    def _print_debug(frame: Frame, received: bool) -> None:
        """This function is used to print the sent/received frames on the log.

        Args:
            frame (Frame): Frame that is being sent/received.
            received (bool): Flag to define if we are sending/receiving.
        """
        msg = "Frame {action}: {frame}".format(action='recv' if received else 'sent', frame=frame)
        logger_sdk.debug(msg)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Time utilities.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

import time


def time_elapsed(start: float) -> float:
    """Return the time elapsed since start.

    Args:
        start (float): Start timestamp.

    Returns:
        float: Time elapsed.
    """
    return time.time() - start



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Event handling.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""


class EventHook:
    """Utility that allows to subscribe multiple callbacks
    to a single event. When the event is emitted the given
    inputs are propagated to all the registered callbacks.

    NOTE: All the callbacks added to the hook need to have
    the same arguments.
    """
    def __init__(self):
        """Class initialization.
        """
        self._handlers = []

    def __iadd__(self, handler: callable) -> 'EventHook':
        """Operator += implementation.
        Adds a function handler to the event hook.

        Args:
            handler (callable): Function handler.

        Returns:
            EventHook: self.
        """
        if handler not in self._handlers:
            self._handlers.append(handler)
        return self

    def __isub__(self, handler: callable) -> 'EventHook':
        """Operator -= implementation.
        Removes a function handler from the event hook.

        Args:
            handler (callable): Function handler.

        Returns:
            EventHook: self.
        """
        if handler not in self._handlers:
            self._handlers.remove(handler)
        return self

    @property
    def empty(self) -> bool:
        """Return if the event hook is empty.

        Returns:
            bool: True if the hook has no handlers.
        """
        return len(self._handlers) == 0

    def clear(self) -> None:
        """Clears the handlers array.
        """
        self._handlers.clear()

    def emit(self, *args, **kwargs) -> None:
        """Emits all the handlers with the given arguments.
        """
        for handler in self._handlers:
            handler(*args, **kwargs)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
SDK threading utilities.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

from threading import Lock, Thread
from typing import Union


class ThreadItem(Thread):
    """Thread implementation.
    All the threads created using this class (daemons or not) can be
    monitored using the ThreadPool class.
    """
    def __init__(self, name: str = 'Unnamed', daemon: bool = True, *args, **kwargs) -> None:
        """Class initialization.

        Args:
            name (str): Thread name.
            daemon (bool): Set the thread as daemon.
        """
        super(ThreadItem, self).__init__(name=name, *args, **kwargs)
        self.setDaemon(daemon)
        self.start()

    def run(self) -> None:
        """Run the thread process.
        """
        ThreadPool().add(thread=self)
        super().run()
        ThreadPool().remove(thread=self)


class ThreadPool(object):
    """Simple thread pool implementation.
    This class is implemented as a singleton.
    """
    MAX_THREADS:    int = 100

    __inst:         Union[None, 'ThreadPool'] = None

    def __new__(cls):
        """Singleton creation.
        """
        if cls.__inst is None:
            cls.__inst = super(ThreadPool, cls).__new__(cls)
        return cls.__inst

    def __init__(self) -> None:
        """Class initialization.
        """
        # Initialize
        self._lock = Lock()
        self._total = 0
        self._details = dict()

    @property
    def total(self) -> int:
        """Return the number of threads being executed.

        Return:
            int: Number of threads.
        """
        with self._lock:
            return self._total

    @property
    def details(self) -> list:
        """Return a list with the quantity and type of the
        threads running.

        Return:
            list: Type/quantity of threads running.
        """
        with self._lock:
            fmt = '{{key:<{size}s}}: {{val:d}}'.format(size=max([len(key) for key in list(self._details.keys())]))
            return [fmt.format(key=key, val=val) for key, val in self._details.items()]

    def add(self, thread: Thread) -> None:
        """Add a thread to the running pool.

        Args:
            thread (Thread): Thread to be added.
        """
        with self._lock:
            # Check the number of threads...
            self._total += 1
            if self._total > self.MAX_THREADS:
                raise ThreadOverflowException()

            # All Ok. Add thread.
            if thread.name in self._details:
                self._details[thread.name] += 1
            else:
                self._details[thread.name] = 1

    def remove(self, thread: Thread) -> None:
        """Remove a thread from the running pool.

        Args:
            thread (Thread): Thread to be removed.
        """
        with self._lock:
            self._total -= 1
            if thread.name in self._details:
                self._details[thread.name] -= 1
                if self._details[thread.name] == 0:
                    self._details.pop(thread.name)


class ThreadOverflowException(Exception):
    """Thread overflow exception.
    """
    def __init__(self):
        super().__init__(
            'Maximum number of threads exceeded ({})'.format(ThreadPool().MAX_THREADS) +
            '\nThread details:\n\t' + '\n\t'.join(ThreadPool().details) + '\n'
            )



#!/usr/bin/python
# -*- coding: ascii -*-
"""
SDK logger.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

import logging
from dataclasses import dataclass


@dataclass
class LoggerFormat:
    """Class used to contain a logger format configuration.

    By default:
    """
    pattern:    str = '{created:.05f}: {name:<8s}: {levelname:<8s}: {module:<20s}: {message:s}'
    style:      str = '{'


class Logger:
    """Implement a basic logger handler class.
    """
    FMT_BASE = LoggerFormat()

    def __init__(self, name: str = '', fmt: LoggerFormat = None) -> None:
        """Class initialization. Apply a basic configuration to the logger.

        Args:
            name (str): Logger name.
            fmt (LoggerFormat): Log entries format.
        """
        # Define formatter
        if fmt is None:
            fmt = self.FMT_BASE

        # Create logger (debug level, disabled)
        self._logger = logging.Logger(name=name)
        self._logger.setLevel(level=logging.DEBUG)
        self._logger.disabled = True

        # Set formatter
        fmt = logging.Formatter(fmt=fmt.pattern, style=fmt.style)
        hdlr = logging.StreamHandler()
        hdlr.setFormatter(fmt=fmt)
        self._logger.addHandler(hdlr=hdlr)

    @property
    def logger(self) -> logging.Logger:
        """Get the logger handler.

        Return:
            logging.Logger: Logger handler.
        """
        return self._logger

    def enable(self) -> None:
        """Enable the logger.
        """
        self._logger.disabled = False

    def disable(self) -> None:
        """Disable the logger.
        """
        self._logger.disabled = True

    def set_level(self, level: int = logging.DEBUG) -> None:
        """Set the logger messages level.

        Args:
            level (int): Logger level.
        """
        self._logger.setLevel(level=level)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Frame implementation.
This classes are used to:
    - Define the command frame payload / serialization.
    - Define how the serial device data should be read to deserialize a frame.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

from abc import abstractmethod
from embutils.serial.core import SerialDevice
from embutils.utils import EventHook, Serialized


class Frame(Serialized):
    """This class define the basic structure of the frame.
    """
    def __repr__(self) -> str:
        """Get the class representation string.

        Return:
            str: Class representation string.
        """
        return f'{self.__class__.__name__}(raw=0x{self.raw().hex()})'

    def __eq__(self, other: 'Frame'):
        """Compare two frames.

        Returns:
            bool: True if equal, false otherwise.
        """
        return self.raw() == other.raw()

    def __ne__(self, other: 'Frame'):
        """Compare two frames.

        Returns:
            bool: True if not equal, false otherwise.
        """
        return not self.__eq__(other=other)

    @abstractmethod
    def raw(self) -> bytearray:
        """Frame contents as bytearray (without encoding).

        Returns:
            bytearray: Frame contents.
        """
        pass


class FrameHandler:
    """This class define how the frame bytes are processed when received.

    The idea is to use this class to define a state machine to receive the frame.
    """
    @abstractmethod
    def read_process(self, serial_device: SerialDevice, emitter: EventHook) -> bool:
        """This method implements the frame reading state machine.

        Args:
            serial_device (bytearray): Device to read the frame bytes.
            emitter (EventHook): Event to be raised when a frame is received.

        Return:
            bool: True if success, false on serial device disconnection or
            reading issues.
        """
        pass



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serial device implementation.
The classes are used to:
    - Abstract the serial device interface.
    - Define/operate over a list of devices.
    - Scan connected devices.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

import serial
import time
from embutils.utils import time_elapsed, EventHook, IntEnumMod, LOG_SDK, ThreadItem, UsbID
from serial.tools import list_ports
from typing import List, Tuple, Union


logger_sdk = LOG_SDK.logger


class SerialDevice:
    """Serial device implementation.
    This class can be used to extend the Serial class capabilities.

    The available events are:
        1. on_usb_id_changed: This event is emitted when, on connection,
            the ID of the serial USB changes.
            Subscribe using callback with syntax:
                def <callback>(old_id: UsbID, new_id: UsbID)
    """
    SETTINGS = {
        'baudrate': 115200,
        'bytesize': 8,
        'stopbits': 1,
        'parity':   'N',
        'timeout':  0.1
        }

    def __init__(self, port: str = None, usb_id: UsbID = None, settings: dict = None, looped: bool = False) -> None:
        """Class initialization.

        Args:
            port (str): Port in which the serial device is connected.
            usb_id (UsbID): Serial device USB ID.
            settings (dict): Serial device configuration.
            looped (bool): Enable the test mode (looped serial).
        """
        # Initialize values
        self._id:       Union[None, UsbID]  = None

        # Check settings
        if not isinstance(settings, dict):
            settings = self.SETTINGS

        # Events
        self.on_usb_id_changed = EventHook()

        # Check if test mode
        if looped:
            # Yes -> Configure looped
            # Ask for ID
            if not isinstance(usb_id, UsbID):
                raise ValueError('Test UsbID is required for looped mode!')
            # Configure
            self._serial = serial.serial_for_url(url='loop://', exclusive=True)
            self._serial.apply_settings(d=settings)
            self._id = usb_id
        else:
            # No -> Configure normal
            # Ask for port
            if not isinstance(port, str):
                msg = 'Serial port is required!'
                logger_sdk.error(msg)
                raise ValueError(msg)
            # Get ID if needed
            if not isinstance(usb_id, UsbID):
                usb_id = self._get_id_from_port(port=port)
                if usb_id is None:
                    msg = 'Serial port is not connected!'
                    logger_sdk.error(msg)
                    raise ValueError(msg)
            # Prepare port
            self._serial = serial.Serial(exclusive=True)
            self._serial.apply_settings(d=settings)
            self.serial.port = port
            self._id = usb_id

        # Log creation
        logger_sdk.info("Device created: port={}, id={}.".format(self.port, self.id))

    def __repr__(self) -> str:
        """Get the class representation string.

        Return:
            str: Class representation string.
        """
        return f'{self.__class__.__name__}(port={self.port}, usb_id={self.id})'

    def __eq__(self, other: 'SerialDevice') -> bool:
        """Compare two serial devices.

        Returns:
            bool: True if equal, false otherwise.
        """
        return (self.port == other.port) and (self.id == other.id)

    @property
    def serial(self) -> serial.Serial:
        """Serial port handler object.

        Returns:
            serial.Serial: Serial interface.
        """
        return self._serial

    @property
    def port(self) -> str:
        """Serial device port.

        Returns:
            str: Serial device port name.
        """
        return self._serial.port

    @property
    def id(self) -> UsbID:
        """Serial device USB ID.

        Returns:
            UsbID: USB ID.
        """
        return self._id

    @property
    def timeout(self) -> float:
        """Serial device operation timeout.

        Return:
            float: Serial device operation timeout.
        """
        return self._serial.timeout

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        """Set the serial device operation timeout.

        Args:
            timeout (float): Serial operation timeout.
        """
        self._serial.timeout = timeout

    @property
    def is_open(self) -> bool:
        """Check if the serial device is open.

        Returns:
            bool: True if open, false otherwise.
        """
        return self._serial.is_open

    def open(self) -> bool:
        """Tries to open the serial device connection.

        Returns:
            bool: True if succeed, false otherwise.
        """
        # Check if port is available
        if self._serial.port is None:
            return False

        # Update USB ID
        usb_id = self._get_id_from_port(port=self._serial.port)
        if usb_id is None:
            logger_sdk.error("Port {} is not connected".format(self.port))
            return False
        elif usb_id != self._id:
            logger_sdk.error("Port {} has changed its ID!".format(self.port))
            ThreadItem(
                name='{}.{}'.format(self.__class__.__name__, 'on_list_change'),
                target=lambda: self.on_usb_id_changed.emit(old_id=self._id, new_id=usb_id)
                )
            self._id = usb_id

        # Try to connect
        try:
            if self._serial.is_open:
                logger_sdk.info("Port {} already open.".format(self.port))
                return True
            else:
                self._serial.open()
                logger_sdk.info("Port {} opened.".format(self.port))
                return True

        except serial.SerialException as ex:
            logger_sdk.error("Port {} couldn't be opened. {}".format(self.port, ex))
            return False

    def close(self) -> None:
        """Close the serial port.
        """
        if self._serial.is_open:
            self.flush()
            self._serial.close()
            logger_sdk.info("Port {} closed.".format(self.port))

    def flush(self) -> None:
        """Flush the serial port buffer.
        """
        try:
            if self._serial.is_open:
                self._serial.flush()
        except serial.SerialException:
            pass

    def write(self, data: bytearray) -> int:
        """Writes a bytearray through the serial port.

        Args:
            data (bytearray): Bytes to be sent.

        Return:
             int: Bytes sent.
        """
        if self._serial.is_open:
            return self._serial.write(data=data)
        return 0

    def read(self, size: int = 1) -> Union[None, bytearray]:
        """Read a given number of bytes form the serial buffer. The process
        is stopped after the timeout is reached.

        Args:
            size (int): Number oo bytes to read.

        Returns:
            Union[None, bytearray]: None if empty or disconnected. Bytearray in case of bytes received.
        """
        try:
            if self._serial.is_open:
                return self._serial.read(size=size)
        except serial.SerialException as ex:
            logger_sdk.error("Port {} has connection issues. {}".format(self.port, ex))
            return None

    def read_until(self, expected: bytes = b'\n', size: int = None) -> Union[None, bytearray]:
        """Read bytes until the expected sequence is found, the bytes exceed the size or a timeout
        is triggered.

        Args:
            expected (bytes): Read target byte.
            size (int): Read size limit.

        Returns:
            Union[None, bytearray]: None if empty or disconnected. Bytearray in case of bytes received.
        """
        try:
            if self._serial.is_open:
                return self._serial.read_until(expected=expected, size=size)
        except serial.SerialException as ex:
            logger_sdk.error("Port {} has connection issues. {}".format(self.port, ex))
            return None

    @staticmethod
    def _get_id_from_port(port: str) -> Union[None, UsbID]:
        dev_list = SerialDeviceList.scan().filter(port=port)
        if dev_list:
            return dev_list[0].id
        return None


class SerialDeviceList(List[SerialDevice]):
    """Serial device list implementation.
    Can be used to list the connected serial devices.
    """
    def get_changes(self, other: 'SerialDeviceList') -> 'SerialDeviceList':
        """Get the differences between two serial device lists.

        Args:
            other (SerialDeviceList): List to compare with.

        Return:
            SerialDeviceList: List containing the filtered devices.
        """
        if self == other:
            return SerialDeviceList()

        # Define base and comparison target
        if len(self) > len(other):
            base = self
            comp = other
        else:
            base = other
            comp = self

        # Perform comparison
        diff = SerialDeviceList()
        for dev in base:
            if dev not in comp:
                diff.append(dev)
        return diff

    def filter(self, port: str = None, usb_id: UsbID = None) -> 'SerialDeviceList':
        """Filter elements on the serial device list.

        Args:
            port (str): Port of interest.
            usb_id (UsbID): Device ID of interest.

        Return:
            SerialDeviceList: List containing filtered devices.
        """
        dev_list = self

        # Filter by port
        if isinstance(port, str):
            dev_list = SerialDeviceList([dev for dev in dev_list if dev.port == port])

        # Filter by ID
        if isinstance(usb_id, UsbID):
            dev_list = SerialDeviceList([dev for dev in dev_list if dev.id == usb_id])

        return dev_list

    @staticmethod
    def scan() -> 'SerialDeviceList':
        """Scan the system for available serial ports.

        Return:
            SerialDeviceList: List including all the connected ports.
        """
        dev_list = SerialDeviceList()

        # Get devices
        logger_sdk.info("Scanning serial devices")
        dev_scan = list_ports.comports()
        for dev in dev_scan:
            # Not consider items without ID
            if dev.vid is None or dev.pid is None:
                continue

            # Create device
            new = SerialDevice(port=dev.device, usb_id=UsbID(vid=dev.vid, pid=dev.pid))
            dev_list.append(new)

        return dev_list


class SerialDeviceScanner:
    """Serial device scanner implementation.
    This class can be used to periodically check the serial devices on the system.

    The available events are:
        1. on_scan_period: This event is emitted after every periodic scan.
            Subscribe using callback with syntax:
                def <callback>()

        2. on_list_change: This event is emitted when a change is detected on the device list.
            Subscribe using callback with syntax:
                def <callback>(event: SerialDeviceScanner.Event, devices: SerialDeviceList)
    """
    class Event(IntEnumMod):
        """Serial device scanner events.
        """
        SD_NO_EVENT             = 0x00      # No event
        SD_LIST_CHANGED         = 0x01      # Serial devices list has changed
        SD_PLUGGED_SINGLE       = 0x02      # A single device was plugged
        SD_PLUGGED_MULTI        = 0x03      # Multiple devices were plugged
        SD_REMOVED_SINGLE       = 0x04      # A single device was unplugged
        SD_REMOVED_MULTI        = 0x05      # Multiple devices were unplugged

        @staticmethod
        def get_event(old: SerialDeviceList, new: SerialDeviceList) -> Tuple['SerialDeviceScanner.Event', SerialDeviceList]:
            """Compare the two lists and return the associates scanner event and difference list.

            Args:
                old (SerialDeviceList): Current device list.
                new (SerialDeviceList): Last scanned list.

            Returns:
                SerialDeviceScanner.Event, SerialDeviceList: Scanner event and difference list.
            """
            # Get difference and define event
            diff = old.get_changes(other=new)
            event = SerialDeviceScanner.Event.SD_NO_EVENT
            if len(new) > len(old):
                # Added devices
                event = event.SD_PLUGGED_MULTI if (len(diff) > 1) else event.SD_PLUGGED_SINGLE
            elif len(old) > len(new):
                # Removed devices
                event = event.SD_REMOVED_MULTI if (len(diff) > 1) else event.SD_REMOVED_SINGLE
            else:
                # Same length but different
                event = event.SD_LIST_CHANGED
            return event, diff

    def __init__(self, scan_period: float = 0.5) -> None:
        """Class initialization.

        Args:
            scan_period (float): Scanner refresh period in seconds.
        """
        # Define device lists
        self._dev_list      = SerialDeviceList()    # Devices connected
        self._dev_change    = SerialDeviceList()    # List of devices that triggered the event

        # Define scan period
        self._scan_period = scan_period

        # Define events
        self.on_scan_period = EventHook()
        self.on_list_change = EventHook()

        # Start scanner
        self._is_active = True
        self._thread = ThreadItem(name=self.__class__.__name__, target=self._process)

    def __del__(self) -> None:
        """Class destructor. Stop the thread.
        """
        self.stop()

    @property
    def is_alive(self) -> bool:
        """Return if the thread is alive.

        Returns:
            bool: True if alive, false otherwise.
        """
        return self._thread.is_alive()

    @property
    def connected_devices(self) -> SerialDeviceList:
        """Return the current scanned device list.

        Returns:
            SerialDeviceList: List with the scanned devices.
        """
        return self._dev_list

    def stop(self) -> None:
        """Stops the serial scanner thread.
        """
        self._is_active = False
        while self._thread.is_alive():
            time.sleep(0.01)
        logger_sdk.info("Scanner stopped.")

    def _scan(self) -> None:
        """This method executed on every period.
        """
        # Get the connected devices
        dev_list = SerialDeviceList.scan()

        # If the change callback has items, inform the differences
        if not self.on_list_change.empty and (dev_list != self._dev_change):
            # Get the changes and generate event
            event, dev_diff = self.Event.get_event(old=self._dev_change, new=dev_list)
            ThreadItem(
                name='{}.{}'.format(self.__class__.__name__, 'on_list_change'),
                target=lambda: self.on_list_change.emit(event=event, dev_diff=dev_diff)
                )

            # Update the compare list
            self._dev_change = dev_list

        # Update the connected devices list
        self._dev_list = dev_list
        ThreadItem(
            name='{}.{}'.format(self.__class__.__name__, 'on_scan_period'),
            target=self.on_scan_period.emit
            )

    def _process(self) -> None:
        """Serial device scanner main process.
        The process consists in:
            1. Read the connected devices.
            2. Compare the new/old lists and generate the change events.
            3. Update connected device lists.
        """
        logger_sdk.info("Scanner started.")

        # Initialize connected devices
        self._dev_list = SerialDeviceList.scan()

        # Execute the first scan
        self._scan()

        # Do this periodically
        last_update = time.time()
        while self._is_active:
            # Wait until the scan period is passed...
            if time_elapsed(last_update) > self._scan_period:
                last_update = time.time()
                self._scan()

            # Let some time pass
            time.sleep(0.01)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Byte manipulation utilities.

@date:      2021
@author:    Christian Wiche
@contact:   cwichel@gmail.com
@license:   The MIT License (MIT)
"""

from math import ceil
from typing import Union


def bitmask(bit: int, fill: bool = False) -> int:
    """Return a bitmask.

    Args:
        bit (int): Bit position.
        fill (bool): Fill all the mask with ones.

    Return:
        int: bitmask.
    """
    mask = (1 << bit)
    return (((mask - 1) << 1) | 1) if fill else mask


def reverse_bits(data: int, size: int = 8) -> int:
    """Reverse the bits of the input data.
    Ex: b10000010 -> b01000001

    Args:
        data (int): Input byte.
        size (int): The size of the output in bits.

    Return:
        int: Reverser value.
    """
    data &= bitmask(bit=size - 1, fill=True)
    aux = '{data:0{size:d}b}'.format(data=data, size=size)
    return int(aux[::-1], base=2)


def reverse_bytes(data: bytearray) -> bytearray:
    """Return the reversed byte array

    Args:
        data (bytearray): Bytes to be reversed.

    Return:
         bytearray: Reversed bytes.
    """
    return data[::-1]


def as_bin(data: int, size: int = 8) -> str:
    """Return the input data as a BIN string.

    Note: The output will be truncated by the size.

    Args:
        data (int): Input value.
        size (int): Size in bits.

    Return:
        str: Binary string.
    """
    data &= bitmask(bit=size - 1, fill=True)
    return '0b{value:0{size:d}b}'.format(value=data, size=size)


def as_hex(data: Union[int, bytearray], size: int = 8) -> str:
    """Return the input data as a HEX string.

    Note: The output will be truncated by the size.

    Args:
        data (int): Input value.
        size (int): Size in bits.

    Return:
        str: Binary string.
    """
    if isinstance(data, bytearray):
        return '0x' + ''.join(['{byte:02X}'.format(byte=byte) for byte in data])

    else:
        data &= bitmask(bit=(size - 1), fill=True)
        return '0x{value:0{size:d}X}'.format(value=data, size=ceil(size / 4))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['embutils',
 'embutils.serial',
 'embutils.serial.core',
 'embutils.serial.data',
 'embutils.utils']

package_data = \
{'': ['*']}

install_requires = \
['pyserial>=3.5,<4.0']

entry_points = \
{'console_scripts': ['test = scripts.poetry:run_test',
                     'version = scripts.poetry:run_version']}

setup_kwargs = {
    'name': 'embutils',
    'version': '0.1.1',
    'description': 'Python utilities for embedded projects',
    'long_description': '# Embutils [![PyPI version](https://badge.fury.io/py/embutils.svg)](https://badge.fury.io/py/embutils) [![License](https://img.shields.io/:license-mit-blue.svg?style=flat-square)](https://badges.mit-license.org)\n\nPython utilities for embedded projects. \n\n## Installation \nYou can get the packaged version from [PyPI](https://pypi.org/project/embutils/):\n```\npip install embutils\n```',
    'author': 'Christian Wiche',
    'author_email': 'cwichel@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/cwichel/embutils',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from .bytes import bitmask, reverse_bits, reverse_bytes, as_bin, as_hex
from .cobs import COBS
from .crc import CRC
from .enum import IntEnumMod
from .events import EventHook
from .logger import LoggerFormat, Logger
from .serialized import Serialized
from .threading import ThreadItem, ThreadPool, ThreadOverflowException
from .time import time_elapsed
from .usb_id import UsbID

# Create a logger for internal use
LOG_SDK = Logger(name='EMBUTILS')
